package cn.com.qmth.print.manage.service.impl;

import cn.com.qmth.print.manage.config.PmConstants;
import cn.com.qmth.print.manage.dao.ExamDao;
import cn.com.qmth.print.manage.dto.ExamDTO;
import cn.com.qmth.print.manage.entity.*;
import cn.com.qmth.print.manage.enums.CheckSort;
import cn.com.qmth.print.manage.enums.GroupType;
import cn.com.qmth.print.manage.enums.RecordStatus;
import cn.com.qmth.print.manage.service.*;
import cn.com.qmth.print.manage.service.query.ExamQuery;
import cn.com.qmth.print.manage.utils.excel.ExcelError;
import cn.com.qmth.print.manage.utils.excel.ExcelReader;
import cn.com.qmth.print.manage.vo.ExamVo;
import com.baomidou.mybatisplus.core.conditions.query.QueryWrapper;
import com.baomidou.mybatisplus.core.metadata.IPage;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import com.baomidou.mybatisplus.extension.service.impl.ServiceImpl;
import com.qmth.boot.core.exception.StatusException;
import org.springframework.beans.BeanUtils;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.transaction.interceptor.TransactionAspectSupport;
import org.springframework.util.CollectionUtils;
import org.springframework.web.multipart.MultipartFile;

import javax.annotation.Resource;
import java.io.IOException;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.time.Year;
import java.util.*;
import java.util.concurrent.atomic.AtomicLong;
import java.util.stream.Collectors;
import java.util.stream.Stream;

@Service
public class ExamServiceImpl extends ServiceImpl<ExamDao, ExamEntity> implements ExamService {

    @Resource
    private ExamUserRelationService examUserRelationService;

    @Resource
    private ExamStudentService examStudentService;

    @Resource
    private CheckRecordService checkRecordService;

    @Resource
    private BreakRecordService breakRecordService;

    @Resource
    private ExamDao examDao;

    @Resource
    private OrgService orgService;

    @Override
    public List<ExamEntity> listByOrgId() {
        QueryWrapper<ExamEntity> queryWrapper = new QueryWrapper<>();
        queryWrapper.lambda().orderByDesc(ExamEntity::getId);
        return this.list(queryWrapper);
    }

    @Transactional
    @Override
    public Long saveExam(ExamEntity exam) {
        // 同机构下的考试批次名称不能重复校验
        String examName = exam.getName();
        Long orgId = exam.getOrgId();
        ExamEntity nameCheck = this.getOne(new QueryWrapper<ExamEntity>().lambda()
                .eq(ExamEntity::getName, examName)
                .eq(ExamEntity::getOrgId, orgId));
        if (Objects.nonNull(nameCheck)) {
            Long id = nameCheck.getId();
            if (!Objects.equals(id, exam.getId())) {
                throw new StatusException(String.format("批次名称重复! 已经存在名称为[%s]的批次", examName));
            }
        }

        if (exam.getId() != null) {
            // 编辑
            ExamEntity examEntity = this.getById(exam.getId());
            if (examEntity == null) {
                throw new StatusException("批次不存在");
            }

            // 删除用户绑定
            examUserRelationService.deleteByExamId(exam.getId());
        }
        exam.setCreateTime(new Date());
        exam.setUpdateTime(new Date());
        this.saveOrUpdate(exam);

        // 保存用户绑定
        String[] userIds = exam.getUserIds().split(",");
        for (String userId : userIds) {
            ExamUserRelationEntity examUserRelationEntity = new ExamUserRelationEntity();
            examUserRelationEntity.setExamId(exam.getId());
            examUserRelationEntity.setUserId(Long.valueOf(userId));
            examUserRelationEntity.setCreateTime(new Date());
            examUserRelationEntity.setUpdateTime(new Date());
            examUserRelationService.save(examUserRelationEntity);
        }
        return exam.getId();
    }

    @Override
    public IPage<ExamVo> pageQuery(ExamQuery query) {
        Page<ExamVo> page = new Page<>(query.getPageNumber(), query.getPageSize());
        long mainQueryS = System.currentTimeMillis();
        IPage<ExamVo> iPage = this.baseMapper.pageExam(page, query);
        System.out.println("主查询时间： " + (System.currentTimeMillis() - mainQueryS));
        for (ExamVo record : iPage.getRecords()) {
            // 考生总量
            record.setStudentCount(examStudentService.count(new QueryWrapper<ExamStudentEntity>().lambda().eq(ExamStudentEntity::getExamId,record.getId())));

            // 绑定用户
            List<ExamUserRelationEntity> examUserRelationEntities = examUserRelationService.listByExamId(record.getId());
            String userIds = null;
            if (!CollectionUtils.isEmpty(examUserRelationEntities)) {
                List<String> ids = examUserRelationEntities.stream().map(m -> String.valueOf(m.getUserId())).collect(Collectors.toList());
                userIds = String.join(",", ids);
            }
            record.setUserIds(userIds);

            // 校验数量
            List<CheckRecordEntity> checkRecordEntities = checkRecordService.listByExamId(record.getId());
            record.setCheckCount(checkRecordEntities.size());
            if (!CollectionUtils.isEmpty(checkRecordEntities)) {
                long checkNormalLCount = checkRecordEntities.stream()
                        .filter(m -> m.getStatus().equals(RecordStatus.NORMAL)).count();
                record.setCheckNormalLCount(Math.toIntExact(checkNormalLCount));

                long checkErrorLCount = checkRecordEntities.stream()
                        .filter(m -> m.getStatus().equals(RecordStatus.ERROR)).count();
                record.setCheckErrorCount(Math.toIntExact(checkErrorLCount));
            }

            // 中断数量
            List<BreakRecordEntity> breakRecordEntities = breakRecordService.listByExamId(record.getId());
            record.setBreakCount(breakRecordEntities.size());
            if (!CollectionUtils.isEmpty(breakRecordEntities)) {
                long breakNormalLCount = breakRecordEntities.stream()
                        .filter(m -> m.getStatus().equals(RecordStatus.NORMAL)).count();
                record.setBreakNormalLCount(Math.toIntExact(breakNormalLCount));
            }

            // 计算进度
            BigDecimal progress = BigDecimal.ZERO;
            int checkCount = record.getCheckCount();
            if (checkCount > 0) {
                progress = new BigDecimal(record.getCheckNormalLCount()).divide(new BigDecimal(record.getCheckCount()), 4, RoundingMode.HALF_UP);
                progress = progress.multiply(new BigDecimal(100));
            }
            record.setProgress(progress);
        }
        System.out.println("总查询时间： " + (System.currentTimeMillis() - mainQueryS));
        return iPage;
    }

    @Override
    public List<ExamVo> listByUserId(Long userId) {
        List<ExamEntity> examEntityList = examDao.listByUserId(userId);
        return examEntityList.stream().flatMap(e -> {
            ExamVo vo = new ExamVo();
            BeanUtils.copyProperties(e, vo);

            // 校验数量
            List<CheckRecordEntity> checkRecordEntities = checkRecordService.listByExamId(vo.getId());
            vo.setCheckCount(checkRecordEntities.size());
            if (!CollectionUtils.isEmpty(checkRecordEntities)) {
                long checkNormalLCount = checkRecordEntities.stream()
                        .filter(m -> m.getStatus().equals(RecordStatus.NORMAL)).count();
                vo.setCheckNormalLCount(Math.toIntExact(checkNormalLCount));

                long checkErrorLCount = checkRecordEntities.stream()
                        .filter(m -> m.getStatus().equals(RecordStatus.ERROR)).count();
                vo.setCheckErrorCount(Math.toIntExact(checkErrorLCount));

                // 计算进度
                BigDecimal progress = BigDecimal.ZERO;
                int checkCount = vo.getCheckCount();
                if (checkCount > 0) {
                    progress = new BigDecimal(vo.getCheckNormalLCount()).divide(new BigDecimal(checkCount), 4, RoundingMode.HALF_UP);
                    progress = progress.multiply(new BigDecimal(100));
                }
                vo.setProgress(progress);
            }
            return Stream.of(vo);
        }).collect(Collectors.toList());
    }

    @Override
    public List<ExcelError> importExams(UserEntity requestUser, Long printLeaderId, MultipartFile file) throws IOException {
        List<String> examNameDatasource = this.list().stream()
                .map(ExamEntity::getName)
                .collect(Collectors.toList());
        // 默认机构
        OrgEntity defaultOrg = orgService.findByCode(PmConstants.DEFAULT_ORG_CODE);

        List<ExamEntity> examEntityList = new ArrayList<>();

        ExcelReader excelReader = new ExcelReader(ExamDTO.class);
        AtomicLong sort = new AtomicLong(1L);
        List<ExcelError> excelErrors = excelReader.reader(file.getInputStream(), obj -> {
            try {
                sort.getAndIncrement();
                ExamDTO dto = (ExamDTO) obj;
                String schoolName = dto.getSchoolName();
                Year year = Year.now();
                String examName = (Long.parseLong(year.toString()) + 1) + schoolName;
                if (examNameDatasource.contains(examName)) {
                    throw new StatusException(String.format("导入不成功,已有学校名称则[%s]", examName));
                }
                ExamEntity examEntity = new ExamEntity();
                examEntity.setOrgId(defaultOrg.getId());
                examEntity.setName(examName);
                examEntity.setEnable(true);
                examEntity.setGroupType(GroupType.COURSE);
                examEntity.setCheckSort(CheckSort.ASC);
                examEntity.setCreateTime(new Date());
                examEntity.setUpdateTime(new Date());
                examEntity.setCreatorId(printLeaderId);
                examEntityList.add(examEntity);
                return null;
            } catch (RuntimeException e) {
                // 手动回滚
                TransactionAspectSupport.currentTransactionStatus().setRollbackOnly();
                ExcelError excelError = new ExcelError();
                excelError.setExcelErrorType(e.getMessage());
                return excelError;
            }
        });
        String errors = errorsString(excelErrors);
        if (errors.length() > 0) {
            throw new StatusException(errors);
        }
        if (!CollectionUtils.isEmpty(examEntityList)) {
            this.saveBatch(examEntityList);
        }
        return excelErrors;
    }

    @Transactional(rollbackFor = Exception.class)
    @Override
    public void deleteExam(Long examId) {
        if (checkRecordService.count(new QueryWrapper<CheckRecordEntity>().lambda()
                .eq(CheckRecordEntity::getExamId, examId)
                .ne(CheckRecordEntity::getStatus, RecordStatus.NONE)) > 0) {
            throw new StatusException("已有校验数据，不可重复导入");
        }
        // 删除考试表id
        this.removeById(examId);
        // 删除考生表
        examStudentService.deleteByExamId(examId);
        // 删除校验记录
        checkRecordService.deleteByExamId(examId);
        // 删除中断记录
        breakRecordService.deleteByExamId(examId);
    }

    /**
     * 拼接导入异常信息
     *
     * @param excelErrors 异常信息
     * @return 异常返回结果
     */
    private String errorsString(List<ExcelError> excelErrors) {
        StringJoiner sj = new StringJoiner(";");
        if (!excelErrors.isEmpty()) {
            int forint = excelErrors.size() < 10 ? excelErrors.size() : 9;
            for (int i = 0; i < forint; i++) {
                ExcelError excelError = excelErrors.get(i);
                sj.add("第" + excelError.getRow() + "行," + excelError.getExcelErrorType());
            }
        }
        return sj.toString();
    }
}